#! /s/std/bin/perl

use POSIX;

$compare_exe = "./compare";
$output_dir  = "./DATA/";
$ckpt_dir    = "./CKPTS/";
$err_file    = "./ERRS/errors";
$sleep_time  = 10; 		# seconds
$TMP_STRING  = ".tmp"; 	# this should be appended to half of the ckpt files


# try to set self up as a daemon
if ( ($pid = fork()) < 0) { 
	exit -1; 
} elsif ($pid != 0) {
	exit 0;	# parent goes bye-bye
}

# child continues
POSIX::setsid();
umask 0;

if ($ARGV[0] eq "-d") { print "debug mode \n"; $DEBUG = true; }

# loop for infinity
MAIN_LOOP: while (true) {

# get the list of files in the checkpoint dir
@files = sort (`ls $ckpt_dir`);
$file_cnt = @files;

# if no files, go to sleep for a while
if (0 == $file_cnt) {
	if ($DEBUG) {print "empty dir\n";}
	sleep $sleep_time;
	next MAIN_LOOP;
} 

# check that dir is ok, should be an even number, 
# bec. it should consist of a bunch of pairs, if not, delete 'em all and loop 
if ($file_cnt % 2 != 0) { 
	# its possible we just got in between the hard link creations
	# give it a second chance
	if ( ! $sec_chance ) { 
		$sec_chance = true;
		sleep (1);
		next MAIN_LOOP;
	}
	print_error ("screwed up directory, delete all $file_cnt files");
	foreach $file (@files) {
		unlink $file;
	}
	next MAIN_LOOP;
} else {
	$sec_chance = false;
} 
	
# this is hopefully the normal path of execution, grab each pair of ckpt files
# and send them to the comparer and then unlink them	 
# since we sorted the directory, each pair should be contiguous in the array
for ($i = 0; $i < $file_cnt; $i = $i + 2) {
	chomp ($file1 = $files[$i]);
	chomp ($file2 = $files[$i+1]); 
	# find the longer one and make sure both are referencing the same file
	next unless ( same_job ($file1, $file2) ); 
	$output_file = "$output_dir$file1";
	$command = "$compare_exe $ckpt_dir$file1 $ckpt_dir$file2 >> $output_file";
	chomp ($time = `date`);
	`echo \"====Comparing ckpts at $time====\" >> $output_file`;
	chomp ($size = `ls -la $ckpt_dir$file1 | awk \'{print \$5\'}`);
	`echo \"size: $size\" >> $output_file`;	
	system($command);
	if ($DEBUG) { print ($command."\n"); }	
	unlink ("$ckpt_dir$file1");
	unlink ("$ckpt_dir$file2");
}	

goto MAIN_LOOP;
}

###### subroutines start here ######

# same_job compares two filenames to see whether they appear to be checkpoint
# files for the same job, they should look like xxx.tmp and xxx
# returns 1 if the filenames are for the same job, 0 otherwise
sub
same_job {
	$file1_len = length ($_[0]);
	$file2_len = length ($_[1]);

	if ($file1_len == $file2_len) {
		print_error("Files are same length, what's going on?");
		return 0;
	}

	if ($file1_len < $file2_len) {
		$shortLen = $file1_len;
		$longer   = $_[1];
		$shorter  = $_[0];
	} else {
		$shortLen = $file2_len;
		$longer   = $_[0];
		$shorter  = $_[1];
	}

	# print "shorter is $shorter, longer is $longer.\n";
	$substring = substr ($longer, 0, $shortLen);
	$tmpstring = substr ($longer, $shortLen);
	# print "comparing $substring and $shorter\n";
	if ($tmpstring ne $TMP_STRING) { return 0 }
	return ($substring eq $shorter); 
}

# print an error message to an error file
sub
print_error {
	`echo \"$_[0]\" >> $err_file`;
	chomp ($_[0]);
	print "$_[0]\n";
}
